#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from xml.etree import ElementTree

from tifffile.tifffile import FileHandle, TiffFile, TiffPage, TiffPageSeries

from opentile.common import NativeTiledPage, Tiler
from opentile.geometry import Size, SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata


class PhilipsTiffTiledPage(NativeTiledPage):
    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        base_shape: Size,
        base_mpp: SizeMm,
        jpeg: Jpeg
    ):
        """OpenTiledPage for Philips Tiff-page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        base_shape: Size
            Size of base level in pyramid.
        base_mpp: SizeMm
            Mpp (um/pixel) for base level in pyramid.
        jpeg: Jpeg
            Jpeg instance to use.
        """
        super().__init__(page, fh)
        self._jpeg = jpeg
        self._base_shape = base_shape
        self._base_mpp = base_mpp
        self._pyramid_index = self._calculate_pyramidal_index(self._base_shape)
        self._mpp = self._calculate_mpp(self._base_mpp)
        self._blank_tile = self._create_blank_tile()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, "
            f"{self._base_shape}, {self._base_mpp}, {self._jpeg})"
        )

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp * 1000

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    @property
    def blank_tile(self) -> bytes:
        """Return blank tile."""
        return self._blank_tile

    def _create_blank_tile(self, luminance: float = 1.0) -> bytes:
        """Create a blank tile from a valid tile. Uses the first found
        valid frame (first frame with non-zero value in databytescounts) and
        fills that image with white.

        Parameters
        ----------
        luminance: float = 1.0
            Luminance for tile, 0 = black, 1 = white.

        Returns
        ----------
        bytes:
            Frame bytes from blank tile.

        """
        # Todo, figure out what color to fill with.
        try:
            # Get first frame in page that is not 0 bytes
            valid_frame_index = next(
                index
                for index, datalength in enumerate(self.page.databytecounts)
                if datalength != 0
            )
        except StopIteration:
            raise ValueError("Could not find valid frame in page.")
        tile = self._read_frame(valid_frame_index)
        if self.page.jpegtables is not None:
            tile = Jpeg.add_jpeg_tables(tile, self.page.jpegtables)
        tile = self._jpeg.fill_frame(tile, luminance)
        return tile

    def _read_frame(self, index: int) -> bytes:
        """Read frame at frame index from page. Return blank tile if tile is
        sparse (length of frame is zero or frame indexis outside length of
        frames)

        Parameters
        ----------
        index: int
            Frame index to read from page.

        Returns
        ----------
        bytes:
            Frame bytes from frame index or blank tile.

        """
        if (
            index >= len(self.page.databytecounts) or
            self.page.databytecounts[index] == 0
        ):
            # Sparse tile
            return self.blank_tile
        return super()._read_frame(index)


CastType = TypeVar('CastType', int, float, str)


class PhilipsMetadata(Metadata):
    _scanner_manufacturer: Optional[str] = None
    _scanner_software_versions: Optional[List[str]] = None
    _scanner_serial_number: Optional[str] = None
    _pixel_spacing: Optional[Tuple[float, float]] = None

    TAGS = [
        'DICOM_PIXEL_SPACING',
        'DICOM_ACQUISITION_DATETIME',
        'DICOM_MANUFACTURER',
        'DICOM_SOFTWARE_VERSIONS',
        'DICOM_DEVICE_SERIAL_NUMBER',
        'DICOM_LOSSY_IMAGE_COMPRESSION_METHOD',
        'DICOM_LOSSY_IMAGE_COMPRESSION_RATIO',
        'DICOM_BITS_ALLOCATED',
        'DICOM_BITS_STORED',
        'DICOM_HIGH_BIT',
        'DICOM_PIXEL_REPRESENTATION'
    ]

    def __init__(
        self,
        tiff_file: TiffFile
    ):
        if tiff_file.philips_metadata is None:
            return

        metadata = ElementTree.fromstring(tiff_file.philips_metadata)
        self._tags: Dict[str, Optional[str]] = {
            tag: None for tag in self.TAGS
        }
        for element in metadata.iter():
            if element.tag == 'Attribute' and element.text is not None:
                name = element.attrib['Name']
                if name in self._tags and self._tags[name] is None:
                    self._tags[name] = element.text

    @property
    def scanner_manufacturer(self) -> Optional[str]:
        return self._tags['DICOM_MANUFACTURER']

    @cached_property
    def scanner_software_versions(self) -> Optional[List[str]]:
        print(self._tags['DICOM_SOFTWARE_VERSIONS'])
        if self._tags['DICOM_SOFTWARE_VERSIONS'] is None:
            return None
        return self._split_and_cast_text(
            self._tags['DICOM_SOFTWARE_VERSIONS'],
            str
        )

    @property
    def scanner_serial_number(self) -> Optional[str]:
        return self._tags['DICOM_DEVICE_SERIAL_NUMBER']

    @cached_property
    def aquisition_datetime(self) -> Optional[datetime]:
        if self._tags['DICOM_ACQUISITION_DATETIME'] is None:
            return None
        try:
            return datetime.strptime(
                self._tags['DICOM_ACQUISITION_DATETIME'],
                r'%Y%m%d%H%M%S.%f'
            )
        except ValueError:
            return None

    @cached_property
    def pixel_spacing(self) -> Optional[Tuple[float, float]]:
        if self._tags['DICOM_PIXEL_SPACING'] is None:
            return None
        return tuple(
            self._split_and_cast_text(
                self._tags['DICOM_PIXEL_SPACING'],
                float
            )[0:2]
        )

    @cached_property
    def properties(self) -> Dict[str, Any]:
        properties: Dict[str, Any] = {}
        if self._tags['DICOM_LOSSY_IMAGE_COMPRESSION_METHOD'] is not None:
            properties['lossy_image_compression_method'] = (
                self._split_and_cast_text(
                    self._tags['DICOM_LOSSY_IMAGE_COMPRESSION_METHOD'],
                    str
                )
            )
        if self._tags['DICOM_LOSSY_IMAGE_COMPRESSION_RATIO'] is not None:
            properties['lossy_image_compression_ratio'] = (
                self._split_and_cast_text(
                    self._tags['DICOM_LOSSY_IMAGE_COMPRESSION_RATIO'],
                    float
                )[0]
            )
        if self._tags['DICOM_BITS_ALLOCATED'] is not None:
            properties['bits_allocated'] = int(
                self._tags['DICOM_BITS_ALLOCATED']
            )
        if self._tags['DICOM_BITS_STORED'] is not None:
            properties['bits_stored'] = int(self._tags['DICOM_BITS_STORED'])
        if self._tags['DICOM_HIGH_BIT'] is not None:
            properties['high_bit'] = int(self._tags['DICOM_HIGH_BIT'])

        if self._tags['DICOM_PIXEL_REPRESENTATION'] is not None:
            properties['pixel_representation'] = (
                self._tags['DICOM_PIXEL_REPRESENTATION']
            )
        return properties

    @staticmethod
    def _split_and_cast_text(
        string: str,
        cast_type: Type[CastType]
    ) -> List[CastType]:
        return [
            cast_type(element) for element in string.replace('"', '').split()
        ]


class PhilipsTiffTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path],
        turbo_path: Optional[Union[str, Path]] = None
    ):
        """Tiler for Philips tiff file.

        Parameters
        ----------
        filepath: Union[str, Path]
            Filepath to a Philips-TiffFile.
        turbo_path: Optional[Union[str, Path]] = None
            Path to turbojpeg (dll or so).
        """
        super().__init__(Path(filepath))
        self._fh = self._tiff_file.filehandle

        self._turbo_path = turbo_path
        self._jpeg = Jpeg(self._turbo_path)

        self._level_series_index = 0
        for series_index, series in enumerate(self.series):
            if self.is_label(series):
                self._label_series_index = series_index
            elif self.is_overview(series):
                self._overview_series_index = series_index
        self._metadata = PhilipsMetadata(self._tiff_file)
        assert self._metadata.pixel_spacing is not None
        self._base_mpp = (
            SizeMm.from_tuple(self._metadata.pixel_spacing) * 1000.0
        )
        self._pages: Dict[Tuple[int, int, int], PhilipsTiffTiledPage] = {}

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_philips

    def get_page(
        self,
        series: int,
        level: int,
        page: int = 0
    ) -> PhilipsTiffTiledPage:
        """Return PhilipsTiffTiledPage for series, level, page."""
        if not (series, level, page) in self._pages:
            self._pages[series, level, page] = PhilipsTiffTiledPage(
                self._get_tiff_page(series, level, page),
                self._fh,
                self.base_size,
                self._base_mpp,
                self._jpeg
            )
        return self._pages[series, level, page]

    @staticmethod
    def is_overview(series: TiffPageSeries) -> bool:
        """Return true if series is a overview series."""
        page = series.pages[0]
        assert isinstance(page, TiffPage)
        return page.description.find('Macro') > -1

    @staticmethod
    def is_label(series: TiffPageSeries) -> bool:
        """Return true if series is a label series."""
        page = series.pages[0]
        assert isinstance(page, TiffPage)
        return page.description.find('Label') > -1

    @staticmethod
    def _get_associated_mpp_from_page(page: TiffPage):
        """Return mpp (um/pixel) for associated image (label or
        macro) from page."""
        pixel_size_start_string = 'pixelsize=('
        pixel_size_start = page.description.find(pixel_size_start_string)
        pixel_size_end = page.description.find(')', pixel_size_start)
        pixel_size_string = page.description[
            pixel_size_start + len(pixel_size_start_string):pixel_size_end
        ]
        pixel_spacing = SizeMm.from_tuple(
            [float(v) for v in pixel_size_string.replace('"', '').split(',')]
        )
        return pixel_spacing / 1000.0



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from functools import cached_property
import io
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union, cast

import numpy as np
from PIL import Image
from tifffile.tifffile import (COMPRESSION, FileHandle, TiffFile, TiffPage,
                               svs_description_metadata)

from opentile.common import NativeTiledPage, OpenTilePage, Tiler
from opentile.geometry import Point, Region, Size, SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata


class SvsStripedPage(OpenTilePage):
    _pyramid_index = 0

    def __init__(self, page: TiffPage, fh: FileHandle, jpeg: Jpeg):
        """OpenTiledPage for jpeg striped Svs page, e.g. overview page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        jpeg: Jpeg
            Jpeg instance to use.

        """
        super().__init__(page, fh)
        self._jpeg = jpeg

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, {self._jpeg}"
        )

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        if tile_position != (0, 0):
            raise ValueError("Non-tiled page, expected tile_position (0, 0)")
        indices = range(len(self.page.dataoffsets))
        scans = (self._read_frame(index) for index in indices)
        jpeg_tables = self.page.jpegtables
        frame = self._jpeg.concatenate_scans(
            scans,
            jpeg_tables,
            True
        )
        return frame

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        return self._jpeg.decode(self.get_tile(tile_position))


class SvsLZWPage(OpenTilePage):
    _pyramid_index = 0

    def __init__(self, page: TiffPage, fh: FileHandle, jpeg: Jpeg):
        """OpenTiledPage for lzw striped Svs page, e.g. label page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        jpeg: Jpeg
            Jpeg instance to use.

        """
        super().__init__(page, fh)
        self._jpeg = jpeg

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, {self._jpeg}"
        )

    @property
    def compression(self) -> COMPRESSION:
        """Return compression of page."""
        return COMPRESSION.JPEG

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        return None

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        return self._jpeg.encode(self.get_decoded_tile(tile_position))

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        if tile_position != (0, 0):
            raise ValueError("Non-tiled page, expected tile_position (0, 0)")

        tile = np.concatenate([
            self._get_row(index)
            for index in range(len(self.page.dataoffsets))
        ], axis=1)
        return np.squeeze(tile)

    def _get_row(self, index: int) -> np.ndarray:
        row = self.page.decode(
            self._read_frame(index),
            index
        )[0]
        assert isinstance(row, np.ndarray)
        return row


class SvsTiledPage(NativeTiledPage):
    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        base_shape: Size,
        base_mpp: SizeMm,
        parent: Optional['SvsTiledPage'] = None
    ):
        """OpenTiledPage for Svs Tiff-page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        base_shape: Size
            Size of base level in pyramid.
        base_mpp: SizeMm
            Mpp (um/pixel) for base level in pyramid.
        parent: Optional['SvsTiledPage'] = None
            Parent SvsTiledPage
        """
        super().__init__(page, fh)
        self._base_shape = base_shape
        self._base_mpp = base_mpp
        self._pyramid_index = self._calculate_pyramidal_index(self._base_shape)
        self._mpp = self._calculate_mpp(self._base_mpp)
        self._parent = parent
        self._right_edge_corrupt, self._bottom_edge_corrupt = (
            self._detect_corrupt_edges()
        )
        self._fixed_tiles: Dict[Point, bytes] = {}

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, "
            f"{self._base_shape}, {self._base_mpp})"
        )

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp / 1000

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    @property
    def right_edge_corrupt(self) -> bool:
        return self._right_edge_corrupt

    @property
    def bottom_edge_corrupt(self) -> bool:
        return self._bottom_edge_corrupt

    def _detect_corrupt_edge(self, edge: Region) -> bool:
        """Return true if tiles at edge are corrupt (any tile has a frame
        that has 0 data length)
        Parameters
        ----------
        edge: Region
            Edge to check for corrupt tiles.

        Returns
        ----------
        bool:
            True if edge contains corrupt tiles.
        """
        for tile in edge.iterate_all():
            frame_index = self._tile_point_to_frame_index(tile)
            if self._page.databytecounts[frame_index] == 0:
                return True
        return False

    def _detect_corrupt_edges(self) -> Tuple[bool, bool]:
        """Returns tuple bool indiciting if right and/or bottom edge of page is
        corrupt. Bottom and right edge tiles in svs pyramid levels > 1 can
        sometimes corrupt. This manifests as tiles with 0-length (easy to
        detect) and tiles with scrambled image data (not so easy detect).

        Returns
        ----------
        Tuple[bool, bool]:
            Tuple indicating if right and/or bottom edge is corrupt.

        """
        if self.pyramid_index == 0:
            return False, False
        right_edge = Region(
            Point(self.tiled_size.width - 1, 0),
            Size(1, self.tiled_size.height - 1)
        )
        right_edge_corrupt = self._detect_corrupt_edge(right_edge)

        bottom_edge = Region(
            Point(0, self.tiled_size.height - 1),
            Size(self.tiled_size.width - 1, 1)
        )
        bottom_edge_corrupt = self._detect_corrupt_edge(bottom_edge)

        return right_edge_corrupt, bottom_edge_corrupt

    def _tile_is_at_right_edge(self, tile_point: Point) -> bool:
        """Return true if tile is at right edge of tiled image."""
        return tile_point.x == self.tiled_size.width - 1

    def _tile_is_at_bottom_edge(self, tile_point: Point) -> bool:
        """Return true if tile is at bottom edge of tiled image."""
        return tile_point.y == self.tiled_size.height - 1

    def _get_scaled_tile(
        self,
        tile_point: Point
    ) -> bytes:
        """Create a tile by downscaling from a lower (higher resolution)
        level.

        Parameters
        ----------
        tile_point: Point
            Position of tile in this pyramid level.

        Returns
        ----------
        bytes:
            Scaled tile bytes.

        """
        if self._parent is None:
            raise ValueError("No parent level to get tiles from")
        scale = int(pow(2, self.pyramid_index - self._parent.pyramid_index))
        scaled_tile_region = Region(tile_point, Size(1, 1)) * scale

        # Get decoded tiles
        decoded_tiles = self._parent.get_decoded_tiles(
            [tile.to_tuple() for tile in scaled_tile_region.iterate_all()]
        )
        image_data = np.zeros(
            (self.tile_size * scale).to_tuple() + (3,),
            dtype=np.uint8
        )
        # Insert decoded_tiles into image_data
        for y in range(scale):
            for x in range(scale):
                image_data_index = y * scale + x
                image_data[
                    y * self.tile_size.width:(y + 1) * self.tile_size.width,
                    x * self.tile_size.height:(x + 1) * self.tile_size.height
                ] = decoded_tiles[image_data_index]

        # Resize image_data using Pillow
        image: Image.Image = Image.fromarray(image_data)
        image = image.resize(
            self.tile_size.to_tuple(),
            resample=Image.Resampling.BILINEAR
        )

        # Return compressed image
        if self.compression == COMPRESSION.JPEG:
            image_format = 'jpeg'
            image_options = {'quality': 95}
        elif self.compression == COMPRESSION.APERIO_JP2000_RGB:
            image_format = 'jpeg2000'
            image_options = {'irreversible': True}
        else:
            raise NotImplementedError("Not supported compression")
        with io.BytesIO() as buffer:
            image.save(buffer, format=image_format, **image_options)
            frame = buffer.getvalue()

        if self.compression == COMPRESSION.APERIO_JP2000_RGB:
            # PIL encodes in jp2, find start of j2k and return from there.
            START_TAGS = bytes([0xFF, 0x4F, 0xFF, 0x51])
            start_index = frame.find(START_TAGS)
            return frame[start_index:]
        return frame

    def _get_fixed_tile(self, tile_point: Point) -> bytes:
        """Get or create a fixed tile inplace for a corrupt tile.

        Parameters
        ----------
        tile_point: Point
            Position of tile to get.

        Returns
        ----------
        bytes:
            Fixed tile bytes.

        """
        if tile_point not in self._fixed_tiles:
            fixed_tile = self._get_scaled_tile(tile_point)
            self._fixed_tiles[tile_point] = fixed_tile
        return self._fixed_tiles[tile_point]

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return image bytes for tile at tile position. If tile is marked as
        corrupt, return a fixed tile. Add color space fix if jpeg compression.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        tile_point = Point.from_tuple(tile_position)
        # Check if tile is corrupted
        tile_corrupt = (
            self.right_edge_corrupt and
            self._tile_is_at_right_edge(tile_point)
        ) or (
            self.bottom_edge_corrupt and
            self._tile_is_at_bottom_edge(tile_point)
        )
        if tile_corrupt:
            return self._get_fixed_tile(tile_point)

        tile = super().get_tile(tile_position)
        if self.compression == COMPRESSION.JPEG:
            tile = Jpeg.add_color_space_fix(tile)
        return tile


class SvsMetadata(Metadata):
    def __init__(
        self,
        page: TiffPage
    ):
        self._svs_metadata = svs_description_metadata(page.description)

    @cached_property
    def magnification(self) -> Optional[float]:
        try:
            return float(self._svs_metadata['AppMag'])
        except (KeyError, ValueError):
            return None

    @cached_property
    def aquisition_datetime(self) -> Optional[datetime]:
        try:
            date = datetime.strptime(
                self._svs_metadata['Date'],
                r'%m/%d/%y'
            )
            time = datetime.strptime(
                self._svs_metadata['Time'],
                r'%H:%M:%S'
            )
        except (KeyError, ValueError):
            return None
        return datetime.combine(date, time.time())

    @property
    def mpp(self) -> float:
        return float(self._svs_metadata['MPP'])


class SvsTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path],
        turbo_path: Optional[Union[str, Path]] = None
    ):
        """Tiler for svs file.

        Parameters
        ----------
        filepath: Union[str, Path]
            Filepath to a svs TiffFile.
        """
        super().__init__(Path(filepath))
        self._fh = self._tiff_file.filehandle
        self._turbo_path = turbo_path
        self._jpeg = Jpeg(self._turbo_path)

        for series_index, series in enumerate(self.series):
            if series.name == 'Baseline':
                self._level_series_index = series_index
            elif series.name == 'Label':
                self._label_series_index = series_index
            elif series.name == 'Macro':
                self._overview_series_index = series_index
        self._pages: Dict[Tuple[int, int, int], OpenTilePage] = {}
        if 'InterColorProfile' in self._tiff_file.pages.first.tags:
            icc_profile = (
                self._tiff_file.pages.first.tags['InterColorProfile'].value
            )
            assert isinstance(icc_profile, bytes) or icc_profile is None
            self._icc_profile = icc_profile
        self._metadata = SvsMetadata(self.base_page)
        self._base_mpp = SizeMm(self._metadata.mpp, self._metadata.mpp)

    @property
    def base_mpp(self) -> SizeMm:
        """Return pixel spacing in um/pixel for base level."""
        return self._base_mpp

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_svs

    def _get_level_page(
        self,
        level: int,
        page: int = 0
    ) -> SvsTiledPage:
        series = self._level_series_index
        if level > 0:
            parent = self.get_page(series, level - 1, page)
            parent = cast(SvsTiledPage, parent)
        else:
            parent = None
        svs_page = SvsTiledPage(
            self._get_tiff_page(series, level, page),
            self._fh,
            self.base_size,
            self.base_mpp,
            parent
        )
        return svs_page

    def get_page(
        self,
        series: int,
        level: int,
        page: int = 0
    ) -> OpenTilePage:
        """Return SvsTiledPage for series, level, page."""
        if not (series, level, page) in self._pages:

            if series == self._overview_series_index:
                svs_page = SvsStripedPage(
                    self._get_tiff_page(series, level, page),
                    self._fh,
                    self._jpeg
                )
            elif series == self._label_series_index:
                svs_page = SvsLZWPage(
                    self._get_tiff_page(series, level, page),
                    self._fh,
                    self._jpeg
                )
            elif series == self._level_series_index:
                svs_page = self._get_level_page(level, page)
            else:
                raise NotImplementedError()

            self._pages[series, level, page] = svs_page
        return self._pages[series, level, page]

    def _get_properties(self) -> Dict[str, Any]:
        """Return dictionary with svs properties."""
        svs_metadata = svs_description_metadata(self.base_page.description)
        magnification = getattr(svs_metadata, 'AppMag', None)
        return {
            'magnification': magnification,
        }



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from ctypes import (POINTER, _Pointer, Structure, byref, c_int, c_short,
                    c_ubyte, c_ulong, c_void_p, cast, cdll,
                    create_string_buffer, memmove, pointer)
from pathlib import Path
from struct import calcsize, unpack
from typing import List, Optional, Tuple, Union

import numpy as np
from turbojpeg import (CUSTOMFILTER, TJFLAG_ACCURATEDCT, TJXOP_NONE,
                       TJXOPT_PERFECT, CroppingRegion, TurboJPEG,
                       split_byte_into_nibbles, tjMCUHeight, tjMCUWidth)


def find_turbojpeg_path() -> Optional[Path]:
    # Only windows installs libraries on strange places
    if os.name != 'nt':
        return None
    try:
        bin_path = Path(os.environ['TURBOJPEG'])
    except KeyError:
        raise ValueError(
            "Enviroment variable 'TURBOJPEG' "
            "needs to be set to turbojpeg bin path."
        )
    if not bin_path.is_dir():
        raise ValueError(
            "Enviroment variable 'TURBOJPEG' "
            "is not set to a directory."
        )
    try:
        dll_file = [
            file for file in bin_path.iterdir()
            if file.is_file()
            and 'turbojpeg' in file.name
            and file.suffix == '.dll'
        ][0]
    except IndexError:
        raise ValueError(
            f'Could not find turbojpeg dll in {bin_path}.'
        )
    return dll_file


class BlankStruct(Structure):
    """Struct to send data to blank_image callback function.

    Parameters
    ----------
    subsample: c_int
        Subsample value of image.
    lum: c_int
        Luminance value to use as background when extending the image.
    """
    _fields_ = [
        ("subsample", c_int),
        ("lum", c_int),
    ]


class BlankTransformStruct(Structure):
    _fields_ = [
        ("r", CroppingRegion),
        ("op", c_int),
        ("options", c_int),
        ("data", POINTER(BlankStruct)),
        ("customFilter", CUSTOMFILTER)
    ]


def get_transform_data(transform_ptr):
    # Cast the content of the transform pointer into a transform structure
    transform = cast(transform_ptr, POINTER(BlankTransformStruct)).contents
    # Cast the content of the callback data pointer in the transform
    # structure to a background structure
    return cast(
        transform.data, POINTER(BlankStruct)
    ).contents


def get_np_coeffs(coeffs_ptr, arrayRegion):
    coeff_array_size = arrayRegion.w * arrayRegion.h
    # Read the coefficients in the pointer as a np array (no copy)
    ArrayType = c_short*coeff_array_size
    array_pointer = cast(coeffs_ptr, POINTER(ArrayType))
    coeffs = np.frombuffer(array_pointer.contents, dtype=np.int16)
    coeffs.shape = (
        arrayRegion.h//8,
        arrayRegion.w//8,
        64
    )
    return coeffs


def blank_image(
    coeffs_ptr: c_void_p,
    arrayRegion: CroppingRegion,
    planeRegion: CroppingRegion,
    componentID: int,
    transformID: int,
    transform_ptr: c_void_p
) -> int:
    """Callback function for filling whole image with background color.

    Parameters
    ----------
    coeffs_ptr: c_void_p
        Pointer to the coefficient array for the callback.
    arrayRegion: CroppingRegion
        The width and height coefficient array and its offset relative to
        the component plane.
    planeRegion: CroppingRegion
        The width and height of the component plane of the coefficient array.
    componentID: int
        The component number (i.e. 0, 1, or 2)
    transformID: int
        The index of the transformation in the array of transformation given to
        the transform function.
    transform_ptr: pointer
        Pointer to the transform structure used for the transformation.

    Returns
    ----------
    int
        CFUNCTYPE function must return an int.
    """
    background_data = get_transform_data(transform_ptr)

    if componentID == 0:
        dc_component = background_data.lum
        subsampling = 0
    else:
        dc_component = 0
        subsampling = background_data.subsample
    coeffs = get_np_coeffs(coeffs_ptr, arrayRegion)
    coeffs[:][:][:] = 0

    for x in range(0, arrayRegion.w//tjMCUWidth[subsampling]):
        for y in range(0, arrayRegion.h//tjMCUHeight[subsampling]):
            coeffs[y][x][0] = dc_component

    return 1


class TurboJPEG_patch(TurboJPEG):
    def __init__(
        self,
        lib_turbojpeg_path: Optional[Union[str, Path]] = None
    ):
        if lib_turbojpeg_path is not None:
            lib_turbojpeg_str_path = str(lib_turbojpeg_path)
        else:
            lib_turbojpeg_str_path = str(
                self._find_turbojpeg()
            )
        super().__init__(lib_turbojpeg_str_path)
        turbo_jpeg = cdll.LoadLibrary(lib_turbojpeg_str_path)
        self.__transform = turbo_jpeg.tjTransform
        self.__transform.argtypes = [
            c_void_p,
            POINTER(c_ubyte),
            c_ulong,
            c_int,
            POINTER(c_void_p),
            POINTER(c_ulong),
            POINTER(BlankTransformStruct),
            c_int
        ]
        self.__transform.restype = c_int

    def fill_image(
        self,
        jpeg_buf: bytes,
        background_luminance: float = 1.0,
    ) -> bytes:
        """Lossless fill jpeg image with background luminance.

        Parameters
        ----------
        jpeg_buf: bytes
            Input jpeg image.
        background_luminance: float
            Luminance level (0 -1 ) to fill background when extending image.
            Default to 1, resulting in white background.

        Returns
        ----------
        List[bytes]
            Filled jpeg images.
        """
        handle = self._init_transform()
        try:
            jpeg_array: np.ndarray = np.frombuffer(jpeg_buf, dtype=np.uint8)
            src_addr = self._getaddr(jpeg_array)
            image_width = c_int()
            image_height = c_int()
            jpeg_subsample = c_int()
            jpeg_colorspace = c_int()

            # Decompress header to get input image size and subsample value
            decompress_header_status = self._decompress_header(
                handle,
                src_addr,
                jpeg_array.size,
                byref(image_width),
                byref(image_height),
                byref(jpeg_subsample),
                byref(jpeg_colorspace)
            )

            if decompress_header_status != 0:
                self._report_error(handle)

            # Use callback to fill in background post-transform
            callback_data = BlankStruct(
                jpeg_subsample,
                self._map_luminance_to_dc_dct_coefficient(
                    jpeg_buf,
                    background_luminance
                )
            )
            callback = CUSTOMFILTER(blank_image)

            # Pointers to output image buffers and buffer size
            dest_array = c_void_p()
            dest_size = c_ulong()
            region = CroppingRegion(0, 0, image_width, image_height)
            crop_transform = BlankTransformStruct(
                region,
                TJXOP_NONE,
                TJXOPT_PERFECT,
                pointer(callback_data),
                callback
            )
            # Do the transforms
            transform_status = self.__transform(
                handle,
                src_addr,
                jpeg_array.size,
                1,
                byref(dest_array),
                byref(dest_size),
                byref(crop_transform),
                TJFLAG_ACCURATEDCT
            )

            # Copy the transform results into python bytes
            dest_buf = create_string_buffer(dest_size.value)
            assert dest_array.value is not None
            memmove(dest_buf, dest_array.value, dest_size.value)

            # Free the output image buffers
            self._free(dest_array)

            if transform_status != 0:
                self._report_error(handle)

            return dest_buf.raw

        finally:
            self._destroy(handle)

    @staticmethod
    def _find_dqt(
        jpeg_data: bytes,
        dqt_index: int
    ) -> Optional[int]:
        """Return byte offset to quantification table with index dqt_index in
        jpeg_data.

        Parameters
        ----------
        jpeg_data: bytes
            Jpeg data containing quantification table(s).
        dqt_index: int
            Index of quantificatin table to find (0 - luminance).

        Returns
        ----------
        Optional[int]
            Byte offset to quantification table, or None if not found.
        """
        offset = 0
        while offset < len(jpeg_data):
            dct_table_offset = jpeg_data[offset:].find(bytes([0xFF, 0xDB]))
            if dct_table_offset == -1:
                break
            dct_table_offset += offset
            dct_table_length = unpack(
                '>H',
                jpeg_data[dct_table_offset+2:dct_table_offset+4]
            )[0]
            dct_table_id_offset = dct_table_offset + 4
            table_index, _ = split_byte_into_nibbles(
                jpeg_data[dct_table_id_offset]
            )
            if table_index == dqt_index:
                return dct_table_offset
            offset += dct_table_offset+dct_table_length
        return None

    @classmethod
    def _get_dc_dqt_element(
        cls,
        jpeg_data: bytes,
        dqt_index: int
    ) -> int:
        """Return dc quantification element from jpeg_data for quantification
        table dqt_index.

        Parameters
        ----------
        jpeg_data: bytes
            Jpeg data containing quantification table(s).
        dqt_index: int
            Index of quantificatin table to get (0 - luminance).

        Returns
        ----------
        int
            Dc quantification element.
        """
        dqt_offset = cls._find_dqt(jpeg_data, dqt_index)
        if dqt_offset is None:
            raise ValueError(
                f"Quantisation table {dqt_index} not found in header"
            )
        precision_offset = dqt_offset+4
        precision = split_byte_into_nibbles(jpeg_data[precision_offset])[0]
        if precision == 0:
            unpack_type = '>b'
        elif precision == 1:
            unpack_type = '>h'
        else:
            raise ValueError('Not valid precision definition in dqt')
        dc_offset = dqt_offset + 5
        dc_length = calcsize(unpack_type)
        dc_value: int = unpack(
            unpack_type,
            jpeg_data[dc_offset:dc_offset+dc_length]
        )[0]
        return dc_value

    @classmethod
    def _map_luminance_to_dc_dct_coefficient(
        cls,
        jpeg_data: bytes,
        luminance: float
    ) -> int:
        """Map a luminance level (0 - 1) to quantified dc dct coefficient.
        Before quantification dct coefficient have a range -1024 - 1023. This
        is reduced upon quantification by the quantification factor. This
        function maps the input luminance level range to the quantified dc dct
        coefficient range.

        Parameters
        ----------
        jpeg_data: bytes
            Jpeg data containing quantification table(s).
        luminance: float
            Luminance level (0 - black, 1 - white).

        Returns
        ----------
        int
            Quantified luminance dc dct coefficent.
        """
        luminance = min(max(luminance, 0), 1)
        dc_dqt_coefficient = cls._get_dc_dqt_element(jpeg_data, 0)
        return(round((luminance * 2047 - 1024) / dc_dqt_coefficient))

    def _find_turbojpeg(self) -> str:
        return self._TurboJPEG__find_turbojpeg()  # type: ignore

    def _init_transform(self) -> c_void_p:
        return self._TurboJPEG__init_transform()  # type: ignore

    def _getaddr(self, nda: np.ndarray) -> _Pointer:
        return self._TurboJPEG__getaddr(nda)  # type: ignore

    def _decompress_header(
        self,
        handle: c_void_p,
        src_addr,
        jpeg_array_size: int,
        image_width,
        image_height,
        jpeg_subsample,
        jpeg_colorspace
    ) -> c_int:
        return self._TurboJPEG__decompress_header(  # type: ignore
            handle,
            src_addr,
            jpeg_array_size,
            image_width,
            image_height,
            jpeg_subsample,
            jpeg_colorspace
        )

    def _report_error(
        self,
        handle: c_void_p
    ) -> None:
        self._TurboJPEG__report_error(handle)  # type: ignore

    def _free(
        self,
        dest_array: c_void_p
    ) -> None:
        self._TurboJPEG__free(dest_array)  # type: ignore

    def _destroy(
        self,
        handle: c_void_p
    ) -> c_int:
        return self._TurboJPEG__destroy(handle)  # type: ignore

    @staticmethod
    def _need_fill_background(
        crop_region: CroppingRegion,
        image_size: Tuple[int, int],
        background_luminance: float
    ) -> bool:
        return TurboJPEG._TurboJPEG__need_fill_background(  # type: ignore
            crop_region,
            image_size,
            background_luminance
        )

    @staticmethod
    def _define_cropping_regions(
        crop_parameters: List[Tuple[int, int, int, int]]
    ) -> List[CroppingRegion]:
        return TurboJPEG._TurboJPEG__define_cropping_regions(  # type: ignore
            crop_parameters
        )



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from struct import pack, unpack
from typing import Iterator, List, Optional, Sequence, Tuple, Union

import numpy as np
from turbojpeg import TJPF_RGB, tjMCUHeight, tjMCUWidth

from opentile.geometry import Size
from opentile.turbojpeg_patch import TurboJPEG_patch as TurboJPEG
from opentile.turbojpeg_patch import find_turbojpeg_path


class JpegTagNotFound(Exception):
    """Raised when expected Jpeg tag was not found."""
    pass


class JpegCropError(Exception):
    """Raised when crop operation fails."""
    pass


class Jpeg:
    TAGS = {
        'tag marker': 0xFF,
        'start of image': 0xD8,
        'application default header': 0xE0,
        'quantization table': 0xDB,
        'start of frame': 0xC0,
        'huffman table': 0xC4,
        'start of scan': 0xDA,
        'end of image': 0xD9,
        'restart interval': 0xDD,
        'restart mark': 0xD0
    }

    def __init__(
        self,
        turbo_path: Optional[Union[str, Path]] = None
    ) -> None:
        if turbo_path is None:
            turbo_path = find_turbojpeg_path()
        self._turbo_jpeg = TurboJPEG(turbo_path)

    def concatenate_fragments(
        self,
        fragments: Iterator[bytes],
        header: bytes
    ) -> bytes:
        """Return frame created by vertically concatenating fragments.

        Parameters
        ----------
        fragments: Iterator[bytes]
            Iterator providing fragments to concatenate.
        header: bytes
            Heaeder for the frame.

        Returns
        ----------
        bytes:
            Concatenated frame in bytes.
        """
        frame = header
        for fragment_index, fragment in enumerate(fragments):
            if not (
                fragment[-2] == Jpeg.TAGS['tag marker']
                and fragment[-1] != b'0'
            ):
                raise JpegTagNotFound(
                    "Tag for end of scan or restart marker not found in scan"
                )
            frame += fragment[:-1]  # Do not include restart mark index
            frame += self.restart_mark(fragment_index)
        frame += self.end_of_image()
        return frame

    def concatenate_scans(
        self,
        scans: Iterator[bytes],
        jpeg_tables: Optional[bytes],
        colorspace_fix: bool = False
    ) -> bytes:
        """Return frame created by horisontal concatenating scans. Scans must
        have the same header content, and only the last scan is allowed to have
        a different heigth.

        Parameters
        ----------
        scans: Iterator[bytes]
            Iterator providing scans to concatenate.
        jpeg_tables: Optional[bytes]
            Optional jpeg tables to insert into frame.
        colorspace_fix: bool = False
            If to apply color space fix (for svs files).

        Returns
        ----------
        bytes:
            Concatenated frame in bytes.
        """
        frame = bytearray()
        image_size: Optional[Size] = None
        scan_size: Optional[Size] = None
        subsample: Optional[int] = None
        for scan_index, scan in enumerate(scans):
            width, height, _subsample, _ = self._turbo_jpeg.decode_header(scan)
            if image_size is None:
                image_size = Size(width, height)
                scan_size = Size(width, height)
                scan_start = 0
                subsample = _subsample
            else:
                image_size.height += height
                start_of_scan, length = self._find_tag(
                    scan,
                    self.start_of_scan()
                )
                if start_of_scan is None or length is None:
                    raise JpegTagNotFound(
                        'Start of scan not found in header'
                    )
                scan_start = start_of_scan + length + 2

            frame += scan[scan_start:-2]
            frame += b'\xFF' + self.restart_mark(scan_index)

        frame[-2:] = self.end_of_image()

        if jpeg_tables is not None:
            frame = self._add_jpeg_tables(frame, jpeg_tables)
        if colorspace_fix:
            frame = self._add_color_space_fix(frame)

        assert (
            image_size is not None
            and scan_size is not None
            and subsample is not None
        )
        frame = self._manipulate_header(
            frame,
            image_size,
            scan_size.area // self.subsample_to_mcu_size(subsample)
        )
        return bytes(frame)

    def fill_frame(self, frame: bytes, luminance: float) -> bytes:
        """Return frame filled with color from luminance.

        Parameters
        ----------
        frame: bytes
            Frame to fill.
        luminance: float
            Luminance to fill (0: black - 1: white).

        Returns
        ----------
        bytes:
            Frame with constant color from luminance.
        """
        return self._turbo_jpeg.fill_image(frame, luminance)

    def decode(self, frame: bytes) -> np.ndarray:
        """Decode frame to np array.

        Parameters
        ----------
        frame: bytes
            Frame to decode.

        Returns
        ----------
        np.ndarray:
            Decoded frame as np array.
        """
        return self._turbo_jpeg.decode(frame, pixel_format=TJPF_RGB)

    def encode(self, data: np.ndarray) -> bytes:
        """Encode np array to bytes.

        Parameters
        ----------
        data: np.ndarray
            Numpy array to encode.

        Returns
        ----------
        bytes:
            Encoded frame of data.
        """
        return self._turbo_jpeg.encode(data, pixel_format=TJPF_RGB)

    def crop_multiple(
        self,
        frame: bytes,
        crop_parameters: Sequence[Tuple[int, int, int, int]]
    ) -> List[bytes]:
        """Crop multipe frames out of frame.

        Parameters
        ----------
        frame: bytes
            Frame to crop from.
        crop_parameters: Sequence[Tuple[int, int, int, int]]
            Parameters for each crop, specified as left position, top position,
            widht, height.

        Returns
        ----------
        List[bytes]:
            Croped frames.
        """
        try:
            return self._turbo_jpeg.crop_multiple(frame, crop_parameters)
        except OSError:
            raise JpegCropError(
                f"Crop of frame failed "
                f"with parameters {crop_parameters}"
            )

    @classmethod
    def add_jpeg_tables(
        cls,
        frame: bytes,
        jpeg_tables: bytes
    ) -> bytes:
        """Add jpeg tables to frame. Tables are insterted before 'start of
        scan'-tag, and leading 'start of image' and ending 'end of image' tags
        are removed from the header prior to insertion.

        Parameters
        ----------
        frame: bytes
            'Abbreviated' jpeg frame lacking jpeg tables.
        jpeg_tables: bytes
            Jpeg tables to add

        Returns
        ----------
        bytes:
            'Interchange' jpeg frame containg jpeg tables.

        """
        return bytes(cls._add_jpeg_tables(bytearray(frame), jpeg_tables))

    @classmethod
    def add_color_space_fix(
        cls,
        frame: bytes
    ) -> bytes:
        """Add color space fix to frame (for svs).

        Parameters
        ----------
        frame: bytes
            Frame to add color space fix to.

        Returns
        ----------
        bytes:
            Frame with color fix.

        """
        return bytes(cls._add_color_space_fix(bytearray(frame)))

    @classmethod
    def manipulate_header(
        cls,
        frame: bytes,
        image_size: Optional[Size] = None,
        restart_interval: Optional[int] = None
    ) -> bytes:
        """Return frame with changed header to reflect changed image size
        or restart interval.

        Parameters
        ----------
        frame: bytes
            Frame with header to update.
        image_size: Optional[Size] = None
            Image size to update header with.
        restart_interval: Optional[int] = None
            Restart interval to update header with.

        Returns
        ----------
        bytes:
            Frame with updated header.

        """
        return bytes(cls._manipulate_header(
            bytearray(frame),
            image_size,
            restart_interval
        ))

    @classmethod
    def start_of_frame(cls) -> bytes:
        """Return bytes representing a start of frame tag."""
        return bytes([cls.TAGS['tag marker'], cls.TAGS['start of frame']])

    @classmethod
    def start_of_scan(cls) -> bytes:
        """Return bytes representing a start of scan tag."""
        return bytes([cls.TAGS['tag marker'], cls.TAGS['start of scan']])

    @classmethod
    def end_of_image(cls) -> bytes:
        """Return bytes representing a end of image tag."""
        return bytes([cls.TAGS['tag marker'], cls.TAGS['end of image']])

    @classmethod
    def restart_mark(cls, index: int) -> bytes:
        """Return bytes representing a restart marker of index (0-7), without
        the prefixing tag (0xFF)."""
        return bytes([cls.TAGS['restart mark'] + index % 8])

    @classmethod
    def restart_interval(cls) -> bytes:
        return bytes([cls.TAGS['tag marker'], cls.TAGS['restart interval']])

    @staticmethod
    def code_short(value: int) -> bytes:
        return pack(">H", value)

    @staticmethod
    def subsample_to_mcu_size(
        subsample: int
    ) -> int:
        return tjMCUWidth[subsample] * tjMCUHeight[subsample]

    @staticmethod
    def _find_tag(
        frame: Union[bytes, bytearray],
        tag: bytes
    ) -> Tuple[Optional[int], Optional[int]]:
        """Return first index and length of payload of tag in header.

        Parameters
        ----------
        frame: bytes
            Frame with header to search.
        tag: bytes
            Tag to search for.

        Returns
        ----------
        Tuple[Optional[int], Optional[int]]:
            Position of tag in header and length of payload.
        """
        index = frame.find(tag)
        if index != -1:
            (length, ) = unpack('>H', frame[index + 2:index + 4])
            return index, length

        return None, None

    @classmethod
    def _manipulate_header(
        cls,
        frame: bytearray,
        size: Optional[Size] = None,
        restart_interval: Optional[int] = None
    ) -> bytearray:
        """Return manipulated header with changed pixel size (width, height)
        and/or restart interval.

        Parameters
        ----------
        frame: bytearray
            Frame to manipulate header of.
        size: Optional[Size] = None
            Pixel size to insert into header.
        restart_interval: Optional[int] = None
            Restart interval to code into header.

        Returns
        ----------
        bytearray:
            Manupulated header.
        """
        if size is not None:
            start_of_frame_index, _ = cls._find_tag(
                frame, cls.start_of_frame()
            )
            if start_of_frame_index is None:
                raise JpegTagNotFound('Start of frame tag not found in header')
            size_index = start_of_frame_index + 5
            frame[size_index:size_index + 2] = cls.code_short(size.height)
            frame[size_index + 2:size_index + 4] = cls.code_short(size.width)

        if restart_interval is not None:
            restart_payload = cls.code_short(restart_interval)
            restart_index, _ = cls._find_tag(frame, cls.restart_interval())
            if restart_index is not None:
                # Modify excisting restart tag
                payload_index = restart_index + 4
                frame[payload_index:payload_index + 2] = restart_payload
            else:
                # Make and insert new restart tag
                start_of_scan_index, _ = cls._find_tag(
                    frame, cls.start_of_scan()
                )
                if start_of_scan_index is None:
                    raise JpegTagNotFound(
                        "Start of scan tag not found in header"
                    )
                frame[start_of_scan_index:start_of_scan_index] = (
                    cls.restart_interval()
                    + cls.code_short(4)
                    + restart_payload
                )
        return frame

    @classmethod
    def _add_jpeg_tables(
        cls,
        frame: bytearray,
        jpegtables: bytes
    ) -> bytearray:
        """Add jpeg tables to frame."""
        start_of_scan = frame.find(cls.start_of_scan())
        frame[start_of_scan:start_of_scan] = jpegtables[2:-2]
        return frame

    @classmethod
    def _add_color_space_fix(
        cls,
        frame: bytearray,
    ) -> bytearray:
        """Add Adobe APP14 marker with transform flag 0 indicating image is
        encoded as RGB (not YCbCr)"""
        start_of_scan = frame.find(cls.start_of_scan())
        frame[start_of_scan:start_of_scan] = (
            b"\xFF\xEE\x00\x0E\x41\x64\x6F\x62"
            + b"\x65\x00\x64\x80\x00\x00\x00\x00"
        )
        return frame



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from functools import cached_property
import math
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
from tifffile.tifffile import COMPRESSION, TIFF, FileHandle, TiffFile, TiffPage

from opentile.common import OpenTilePage, Tiler
from opentile.geometry import Point, Region, Size, SizeMm
from opentile.jpeg import Jpeg, JpegCropError
from opentile.metadata import Metadata


def get_value_from_ndpi_comments(
    comments: str,
    value_name: str,
    value_type: Any
) -> Any:
    """Read value from ndpi comment string."""
    for line in comments.split("\n"):
        if value_name in line:
            value_string = line.split('=')[1]
            return value_type(value_string)


class NdpiCache():
    """Cache for bytes ordered by tile position. Oldest entry is removed when
    size of content is above set size."""
    def __init__(self, size: int):
        """Create cache for size items.

        Parameters
        ----------
        size: int
            Size of the cache.

        """
        self._size = size
        self._content: Dict[Point, bytes] = {}
        self._history: List[Point] = []

    def __len__(self) -> int:
        return len(self._history)

    def __str__(self) -> str:
        return (
            f"{type(self).__name__} of size {len(self)} "
            f"and max size {self._size}"
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._size})"

    def __setitem__(self, key: Point, value: bytes) -> None:
        """Set item in cache. Remove old items if needed.

        Parameters
        ----------
        key: Point
            Key for item to set.
        value: bytes:
            Value for item to set.

        """
        self._content[key] = value
        self._history.append(key)
        self._remove_old()

    def __getitem__(self, key: Point) -> bytes:
        """Get item from cache.

        Parameters
        ----------
        key: Point
            Key for item to get.

        Returns
        ----------
        bytes
            Value for key.
        """
        return self._content[key]

    def __contains__(self, key: Point) -> bool:
        """Return true if key in cache.

        Parameters
        ----------
        key: Point
            Key to check for.

        Returns
        ----------
        bool
            True if key is in cache.
        """
        return key in self._content

    @property
    def size(self) -> int:
        return self._size

    def update(self, items: Dict[Point, bytes]) -> None:
        """Update items in cache. Remove old items if needed.

        Parameters
        ----------
        items: Dict[Point, bytes]
            Items to update.

        """
        self._content.update(items)
        self._history += list(items.keys())
        self._remove_old()

    def _remove_old(self) -> None:
        """Remove old items in cache if needed."""
        while len(self._history) > self._size:
            key_to_remove = self._history.pop(0)
            self._content.pop(key_to_remove)


class NdpiTile:
    """Defines a tile by position and coordinates and size for cropping out
    out frame."""
    def __init__(
        self,
        position: Point,
        tile_size: Size,
        frame_size: Size
    ) -> None:
        """Create a ndpi tile and calculate cropping parameters.

        Parameters
        ----------
        position: Point
            Tile position.
        tile_size: Size
            Tile size.
        frame_size: Size
            Frame size.

        """
        self._position = position
        self._tile_size = tile_size
        self._frame_size = frame_size

        self._tiles_per_frame = Size.max(
            self._frame_size // self._tile_size,
            Size(1, 1)
        )
        position_inside_frame: Point = (
            (self.position * self._tile_size)
            % Size.max(self._frame_size, self._tile_size)
        )
        self._left = position_inside_frame.x
        self._top = position_inside_frame.y
        self._frame_position = (
            (self.position // self._tiles_per_frame) * self._tiles_per_frame
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NdpiTile):
            return (
                self.position == other.position and
                self._tile_size == other._tile_size and
                self._frame_size == other._frame_size
            )
        return NotImplemented

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.position}, {self._tile_size}, "
            f"{self._frame_size})"
        )

    def __str__(self) -> str:
        return f"{type(self).__name__} of position {self.position}"

    @property
    def position(self) -> Point:
        """Return position of tile."""
        return self._position

    @property
    def frame_position(self) -> Point:
        """Return frame position for tile."""
        return self._frame_position

    @property
    def left(self) -> int:
        """Return left coordinate for tile inside frame."""
        return self._left

    @property
    def top(self) -> int:
        """Return top coordinate for tile inside frame."""
        return self._top

    @property
    def width(self) -> int:
        """Return width for tile inside frame."""
        return self._tile_size.width

    @property
    def height(self) -> int:
        """Return height for tile inside frame."""
        return self._tile_size.height

    @property
    def frame_size(self) -> Size:
        """Return frame size."""
        return self._frame_size


class NdpiFrameJob:
    """A list of tiles to create from a frame. Tiles need to have the same
    frame position."""
    def __init__(
        self,
        tiles: Union[NdpiTile, List[NdpiTile]]
    ) -> None:
        """Create a frame job from given tile(s).

        Parameters
        ----------
        tiles: Union[NdpiTile, List[NdpiTile]]
            Tile(s) to base the frame job on.

        """
        if isinstance(tiles, NdpiTile):
            tiles = [tiles]
        first_tile = tiles.pop(0)
        self._position = first_tile.frame_position
        self._frame_size = first_tile.frame_size
        self._tiles: List[NdpiTile] = [first_tile]
        for tile in tiles:
            self.append(tile)

    def __repr__(self) -> str:
        return f"{type(self).__name__}{self.tiles})"

    def __str__(self) -> str:
        return f"{type(self).__name__} of tiles {self.tiles}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NdpiFrameJob):
            return self.tiles == other.tiles
        return NotImplemented

    @property
    def position(self) -> Point:
        """The frame position of the frame job."""
        return self._position

    @property
    def frame_size(self) -> Size:
        """Frame size required for reading tiles in NdpiFrameJob."""
        return self._frame_size

    @property
    def tiles(self) -> List[NdpiTile]:
        """Tiles in NdpiFrameJob."""
        return self._tiles

    @property
    def crop_parameters(self) -> List[Tuple[int, int, int, int]]:
        """Parameters for croping tiles from frame in NdpiFrameJob."""
        return [
            (tile.left, tile.top, tile.width, tile.height)
            for tile in self._tiles
        ]

    def append(self, tile: NdpiTile) -> None:
        """Add a tile to the tile job."""
        if tile.frame_position != self.position:
            raise ValueError(f"{tile} does not match {self} frame position")
        self._tiles.append(tile)


class NdpiPage(OpenTilePage):
    _pyramid_index = 0

    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        jpeg: Jpeg
    ):
        """Ndpi page that should not be tiled (e.g. overview or label).
        Image data is assumed to be jpeg.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        jpeg: Jpeg
            Jpeg instance to use.
        """
        super().__init__(page, fh)
        if self.compression != COMPRESSION.JPEG:
            raise NotImplementedError(
                f'{self.compression} is unsupported for ndpi '
                '(Only jpeg is supported)'
            )
        self._jpeg = jpeg
        try:
            # Defined in nm
            assert isinstance(page.ndpi_tags, dict)
            self._focal_plane = (
                page.ndpi_tags['ZOffsetFromSlideCenter'] / 1000.0
            )
        except KeyError:
            self._focal_plane = 0.0

        self._mpp = self._get_mpp_from_page()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, {self._jpeg}"
        )

    @property
    def focal_plane(self) -> float:
        """Return focal plane (in um)."""
        return self._focal_plane

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp / 1000.0

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Return tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        if tile_position != (0, 0):
            raise ValueError("Non-tiled page, expected tile_position (0, 0)")
        return self._read_frame(0)

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        tile = self.get_tile(tile_position)
        return self._jpeg.decode(tile)

    def _get_mpp_from_page(self) -> SizeMm:
        """Return pixel spacing in um/pixel."""
        x_resolution = self.page.tags['XResolution'].value[0]
        y_resolution = self.page.tags['YResolution'].value[0]
        resolution_unit = self.page.tags['ResolutionUnit'].value
        if resolution_unit != TIFF.RESUNIT.CENTIMETER:
            raise ValueError("Unkown resolution unit")
        # 10*1000 um per cm
        mpp_x = 10 * 1000 / x_resolution
        mpp_y = 10 * 1000 / y_resolution
        return SizeMm(mpp_x, mpp_y)


class NdpiTiledPage(NdpiPage, metaclass=ABCMeta):
    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        base_shape: Size,
        tile_size: Size,
        jpeg: Jpeg,
        frame_cache: int = 1
    ):
        """Metaclass for a tiled ndpi page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        base_shape: Size
            Size of base level in pyramid.
        tile_size: Size
            Requested tile size.
        jpeg: Jpeg
            Jpeg instance to use.
        frame_cache: int:
            Number of read frames to cache.
        """
        super().__init__(page, fh, jpeg)
        self._base_shape = base_shape
        self._tile_size = tile_size
        self._file_frame_size = self._get_file_frame_size()
        self._frame_size = Size.max(self.tile_size, self._file_frame_size)
        self._pyramid_index = self._calculate_pyramidal_index(self._base_shape)
        self._frame_cache = NdpiCache(frame_cache)
        self._headers: Dict[Size, bytes] = {}

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, "
            f"{self._base_shape}, {self.tile_size}, {self._jpeg}, "
            f"{self._frame_cache.size})"
        )

    @property
    def suggested_minimum_chunk_size(self) -> int:
        return max(self._frame_size.width // self._tile_size.width, 1)

    @property
    def tile_size(self) -> Size:
        """The size of the tiles to generate."""
        return self._tile_size

    @property
    def frame_size(self) -> Size:
        """The default read size used for reading frames."""
        return self._frame_size

    @property
    def mcu(self) -> Size:
        subsampling: Optional[Tuple[int, int]] = self._page.subsampling
        if subsampling is None or subsampling == (1, 1):
            return Size(8, 8)
        elif subsampling == (2, 1):
            return Size(16, 8)
        elif subsampling == (2, 2):
            return Size(16, 16)
        raise ValueError(f"Unkown subsampling {subsampling}")

    @abstractmethod
    def _read_extended_frame(
        self,
        position: Point,
        frame_size: Size
    ) -> bytes:
        """Read a frame of size frame_size covering position."""
        raise NotImplementedError

    @abstractmethod
    def _get_file_frame_size(self) -> Size:
        """Return size of single frame/stripe in file."""
        raise NotImplementedError

    @abstractmethod
    def _get_frame_size_for_tile(self, tile_position: Point) -> Size:
        """Return frame size used for creating tile at tile position."""
        raise NotImplementedError

    def get_tile(
        self,
        tile_position: Tuple[int, int]
    ) -> bytes:
        """Return image bytes for tile at tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        return self.get_tiles([tile_position])[0]

    def get_tiles(
        self,
        tile_positions: Sequence[Tuple[int, int]]
    ) -> List[bytes]:
        """Return list of image bytes for tile positions.

        Parameters
        ----------
        tile_positions: Sequence[Tuple[int, int]]
            Tile positions to get.

        Returns
        ----------
        List[bytes]
            List of tile bytes.
        """
        frame_jobs = self._sort_into_frame_jobs(tile_positions)
        return [
            tile
            for frame_job in frame_jobs
            for tile in self._create_tiles(frame_job).values()
        ]

    def get_decoded_tiles(
        self, tile_positions: Sequence[Tuple[int, int]]
    ) -> List[np.ndarray]:
        """Return list of decoded tiles for tiles at tile positions.

        Parameters
        ----------
        tile_positions: Sequence[Tuple[int, int]]
            Tile positions to get.

        Returns
        ----------
        List[np.ndarray]
            List of decoded tiles.
        """
        frame_jobs = self._sort_into_frame_jobs(tile_positions)
        return [
            self._jpeg.decode(tile)
            for frame_job in frame_jobs
            for tile in self._create_tiles(frame_job).values()
        ]

    def _create_tiles(
        self,
        frame_job: NdpiFrameJob
    ) -> Dict[Point, bytes]:
        """Return tiles defined by frame job. Read frames are cached by
        frame position.

        Parameters
        ----------
        frame_job: NdpiFrameJob
            Tile job containing tiles that should be created.

        Returns
        ----------
        Dict[Point, bytes]:
            Created tiles ordered by tile coordinate.
        """
        if frame_job.position in self._frame_cache:
            frame = self._frame_cache[frame_job.position]
        else:
            frame = self._read_extended_frame(
                frame_job.position,
                frame_job.frame_size
            )
            self._frame_cache[frame_job.position] = frame
        tiles = self._crop_to_tiles(frame_job, frame)
        return tiles

    def _crop_to_tiles(
        self,
        frame_job: NdpiFrameJob,
        frame: bytes
    ) -> Dict[Point, bytes]:
        """Crop jpeg data to tiles.

        Parameters
        ----------
        frame_job: NdpiFrameJob
            Frame job defining the tiles to produce by cropping jpeg data.
        frame: bytes
            Data to crop from.

        Returns
        ----------
        Dict[Point, bytes]:
            Created tiles ordered by tile coordinate.
        """
        try:
            tiles: List[bytes] = self._jpeg.crop_multiple(
                frame,
                frame_job.crop_parameters
            )
        except JpegCropError:
            raise ValueError(
                f'Failed to crop at position {frame_job.position} with '
                f'parameters {frame_job.crop_parameters}. '
                'This might be due using libjpeg-turbo < 2.1.'
            )
        return {
            tile.position: tiles[i]
            for i, tile in enumerate(frame_job.tiles)
        }

    def _sort_into_frame_jobs(
        self,
        tile_positions: Sequence[Tuple[int, int]]
    ) -> List[NdpiFrameJob]:
        """Sorts tile positions into frame jobs (i.e. from the same frame.)

        Parameters
        ----------
        tile_positions: Sequence[Point]
            List of position to sort.

        Returns
        ----------
        List[NdpiFrameJob]
            List of created frame jobs.

        """
        frame_jobs: Dict[Point, NdpiFrameJob] = {}
        for tile_position in tile_positions:
            tile_point = Point.from_tuple(tile_position)
            if not self._check_if_tile_inside_image(tile_point):
                raise ValueError(
                    f"Tile {tile_point} is outside "
                    f"tiled size {self.tiled_size}"
                )
            frame_size = self._get_frame_size_for_tile(tile_point)
            tile = NdpiTile(tile_point, self.tile_size, frame_size)
            if tile.frame_position in frame_jobs:
                frame_jobs[tile.frame_position].append(tile)
            else:
                frame_jobs[tile.frame_position] = NdpiFrameJob(tile)
        return list(frame_jobs.values())


class NdpiOneFramePage(NdpiTiledPage):
    """Class for a ndpi page containing only one frame that should be tiled.
    The frame can be of any size (smaller or larger than the wanted tile size).
    The frame is padded to an even multipe of tile size.
    """

    def _get_file_frame_size(self) -> Size:
        """Return size of the single frame in file. For single framed page
        this is equal to the level size.

        Returns
        ----------
        Size
            The size of frame in the file.
        """
        return self.image_size

    def _get_frame_size_for_tile(self, tile_position: Point) -> Size:
        """Return read frame size for tile position. For single frame page
        the read frame size is the image size rounded up to the closest tile
        size.

        Returns
        ----------
        Size
            The read frame size.
        """
        return ((self.frame_size) // self.tile_size + 1) * self.tile_size

    def _read_extended_frame(
        self,
        position: Point,
        frame_size: Size
    ) -> bytes:
        """Return padded image covering tile coordinate as valid jpeg bytes.

        Parameters
        ----------
        frame_position: Point
            Upper left tile position that should be covered by the frame.
        frame_size: Size
            Size of the frame to read.

        Returns
        ----------
        bytes
            Frame
        """
        if position != Point(0, 0):
            raise ValueError("Frame position not (0, 0) for one frame level.")
        frame = self._read_frame(0)
        if (
            self.image_size.width % self.mcu.width != 0
            or self.image_size.height % self.mcu.height != 0
        ):
            # Extend to whole MCUs
            even_size = Size.ceil_div(self.image_size, self.mcu) * self.mcu
            frame = Jpeg.manipulate_header(frame, even_size)
        # Use crop_multiple as it allows extending frame
        tile: bytes = self._jpeg.crop_multiple(
            frame,
            [(0, 0, frame_size.width, frame_size.height)]
        )[0]
        return tile


class NdpiStripedPage(NdpiTiledPage):
    """Class for a ndpi page containing stripes. Frames are constructed by
    concatenating multiple stripes, and from the frame one or more tiles can be
    produced by lossless cropping.
    """
    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        base_shape: Size,
        tile_size: Size,
        jpeg: Jpeg,
        frame_cache: int = 1
    ):
        """Ndpi page with striped image data.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandle
            Filehandler to read data from.
        base_shape: Size
            Size of base level in pyramid.
        tile_size: Size
            Requested tile size.
        jpeg: Jpeg
            Jpeg instance to use.
        frame_cache: int:
            Number of read frames to cache.
        """
        super().__init__(page, fh, base_shape, tile_size, jpeg, frame_cache)
        self._striped_size = Size(self.page.chunked[1], self.page.chunked[0])
        jpeg_header = self.page.jpegheader
        assert isinstance(jpeg_header, bytes)
        self._jpeg_header = jpeg_header

    @property
    def stripe_size(self) -> Size:
        """Size of stripes."""
        return self._file_frame_size

    @property
    def striped_size(self) -> Size:
        """Number of stripes in columns and rows."""
        return self._striped_size

    @property
    def jpeg_header(self) -> bytes:
        """Jpeg header in page."""
        return self._jpeg_header

    def _get_file_frame_size(self) -> Size:
        """Return size of stripes in file. For striped levels this is parsed
        from the jpeg header.

        Returns
        ----------
        Size
            The size of stripes in the file.
        """
        stripe_height, stripe_width, _ = self.page.chunks
        return Size(stripe_width, stripe_height)

    def _is_partial_frame(self, tile_position: Point) -> Tuple[bool, bool]:
        """Return a tuple of bools, that are true if tile position is at the
        edge of the image in x or y.

        Parameters
        ----------
        tile_position: int
            Tile position (x or y) to check.

        Returns
        ----------
        Tuple[bool, bool]
            Tuple that is True if tile position x or y is at edge of image.
        """
        partial_x = (
            tile_position.x == (self.tiled_size.width - 1) and
            self.stripe_size.width < self.tile_size.width
        )
        partial_y = (
            tile_position.y == (self.tiled_size.height - 1) and
            self.stripe_size.height < self.tile_size.height
        )
        return partial_x, partial_y

    def _get_frame_size_for_tile(self, tile_position: Point) -> Size:
        """Return frame size used for creating tile at tile position.
        If tile is an edge tile, ensure that the frame does not extend beyond
        the image limits.

        Parameters
        ----------
        tile_position: Point
            Tile position for frame size calculation.

        Returns
        ----------
        Size
            Frame size to be used at tile position.
        """
        is_partial_x, is_partial_y = self._is_partial_frame(tile_position)
        if is_partial_x:
            width = (
                self.stripe_size.width * self.striped_size.width
                - tile_position.x * self.tile_size.width
            )
        else:
            width = self.frame_size.width

        if is_partial_y:
            height = (
                self.stripe_size.height * self.striped_size.height
                - tile_position.y * self.tile_size.height
            )
        else:
            height = self.frame_size.height
        return Size(width, height)

    def _read_extended_frame(
        self,
        position: Point,
        frame_size: Size
    ) -> bytes:
        """Return extended frame of frame size starting at frame position.
        Returned frame is jpeg bytes including header with correct image size.
        Original restart markers are updated to get the proper incrementation.
        End of image tag is appended.

        Parameters
        ----------
        position: Point
            Upper left tile position that should be covered by the frame.
        frame_size: Size
            Size of the frame to get.

        Returns
        ----------
        bytes
            Concatenated frame as jpeg bytes.
        """
        if frame_size in self._headers:
            header = self._headers[frame_size]
        else:
            header = self._jpeg.manipulate_header(
                self.jpeg_header,
                frame_size
            )
            self._headers[frame_size] = header

        stripe_region = Region(
            (position * self.tile_size) // self.stripe_size,
            Size.max(frame_size // self.stripe_size, Size(1, 1))
        )
        indices = [
            self._get_stripe_position_to_index(stripe_coordinate)
            for stripe_coordinate in stripe_region.iterate_all()
        ]
        frame = self._jpeg.concatenate_fragments(
            (self._read_frame(index) for index in indices),
            header
        )
        return frame

    def _get_stripe_position_to_index(self, position: Point) -> int:
        """Return stripe index from position.

        Parameters
        ----------
        position: Point
            position of stripe to get index for.

        Returns
        ----------
        int
            Stripe index.
        """
        return position.x + position.y * self.striped_size.width


class NdpiMetadata(Metadata):
    def __init__(
        self,
        page: TiffPage
    ):
        self._tags = page.tags
        if page.ndpi_tags is not None:
            self._ndpi_tags = page.ndpi_tags
        else:
            self._ndpi_tags = {}

    @cached_property
    def magnification(self) -> Optional[float]:
        try:
            return float(self._ndpi_tags['Magnification'])
        except (AttributeError, ValueError):
            return None

    @property
    def scanner_manufacturer(self) -> Optional[str]:
        return self._ndpi_tags.get('Make')

    @property
    def scanner_model(self) -> Optional[str]:
        return self._ndpi_tags.get('Model')

    @property
    def scanner_software_versions(self) -> Optional[List[str]]:
        software_version = self._ndpi_tags.get('Software')
        if software_version is not None:
            return [software_version]
        return None

    @property
    def scanner_serial_number(self) -> Optional[str]:
        return self._ndpi_tags.get('ScannerSerialNumber')

    @cached_property
    def aquisition_datetime(self) -> Optional[datetime]:
        datetime_tag = self._tags.get('DateTime')
        if datetime_tag is None:
            return None
        try:
            return datetime.strptime(
                datetime_tag.value,
                '%Y:%m:%d %H:%M:%S'
            )
        except ValueError:
            return None

    @cached_property
    def properties(self) -> Dict[str, Any]:
        x_offset_from_slide_center = self._ndpi_tags.get(
            'XOffsetFromSlideCenter'
        )
        y_offset_from_slide_center = self._ndpi_tags.get(
            'YOffsetFromSlideCenter'
        )
        z_offset_from_slide_center = self._ndpi_tags.get(
            'ZXOffsetFromSlideCenter'
        )
        return {
            'x_offset_from_slide_center': x_offset_from_slide_center,
            'y_offset_from_slide_center': y_offset_from_slide_center,
            'z_offset_from_slide_center': z_offset_from_slide_center
        }


class NdpiTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path],
        tile_size: int,
        turbo_path: Optional[Union[str, Path]] = None
    ):
        """Tiler for ndpi file, with functions to produce tiles of specified
        size.

        Parameters
        ----------
        filepath: Union[str, Path]
            Filepath to a ndpi TiffFile.
        tile_size: int
            Tile size to cache and produce. Must be multiple of 8 and will be
            adjusted to be an even multipler or divider of the smallest strip
            width in the file.
        turbo_path: Optional[Union[str, Path]] = None
            Path to turbojpeg (dll or so).

        """
        super().__init__(Path(filepath))

        self._fh = self._tiff_file.filehandle
        self._tile_size = Size(tile_size, tile_size)
        self._tile_size = self._adjust_tile_size(
            tile_size,
            self._get_smallest_stripe_width()
        )
        if self.tile_size.width % 8 != 0 or self.tile_size.height % 8 != 0:
            raise ValueError(f"Tile size {self.tile_size} not divisable by 8")
        self._turbo_path = turbo_path
        self._jpeg = Jpeg(self._turbo_path)

        self._level_series_index = 0
        for series_index, series in enumerate(self.series):
            if series.name == 'Label':
                self._label_series_index = series_index
            elif series.name == 'Macro':
                self._overview_series_index = series_index
        self._pages: Dict[Tuple[int, int, int], NdpiPage] = {}
        self._metadata = NdpiMetadata(self.base_page)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._tiff_file.filename}, "
            f"{self.tile_size.to_tuple}, "
            f"{self._turbo_path})"
        )

    def __str__(self) -> str:
        return f"{type(self).__name__} of Tifffile {self._tiff_file}"

    @property
    def tile_size(self) -> Size:
        """The size of the tiles to generate."""
        return self._tile_size

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return tiff_file.is_ndpi

    def get_page(
        self,
        series: int,
        level: int,
        page: int
    ) -> NdpiPage:
        """Return NdpiPage for series, level, page. NdpiPages holds a cache, so
        store created pages.
        """
        if not (series, level, page) in self._pages:
            ndpi_page = self._create_page(series, level, page)
            self._pages[series, level, page] = ndpi_page
        return self._pages[series, level, page]

    @staticmethod
    def _adjust_tile_size(
        requested_tile_width: int,
        smallest_stripe_width: Optional[int] = None
    ) -> Size:
        """Return adjusted tile size. If file contains striped pages the
        tile size must be an n * smallest stripe width in the file, where n
        is the closest square factor of the ratio between requested tile width
        and smallest stripe width.

        Parameters
        ----------
        requested_tile_width: int
            Requested tile width.
        smallest_stripe_width: Optional[int] = None
            Smallest stripe width in file.

        Returns
        ----------
        Size
            Adjusted tile size.
        """
        if (
            smallest_stripe_width is None or
            smallest_stripe_width == requested_tile_width
        ):
            # No striped pages or requested is equald to smallest
            return Size(requested_tile_width, requested_tile_width)

        if requested_tile_width > smallest_stripe_width:
            factor = requested_tile_width / smallest_stripe_width
        else:
            factor = smallest_stripe_width / requested_tile_width
        # Factor should be a square number (in the series 2^n)
        factor_2 = pow(2, round(math.log2(factor)))
        adjusted_width = factor_2 * smallest_stripe_width
        return Size(adjusted_width, adjusted_width)

    def _get_smallest_stripe_width(self) -> Optional[int]:
        """Return smallest stripe width in file, or None if no page in the
        file is striped.

        Returns
        ----------
        Optional[int]
            The smallest stripe width in the file, or None if no page in the
            file is striped.
        """
        smallest_stripe_width: Optional[int] = None
        for page in self._tiff_file.pages:
            assert isinstance(page, TiffPage)
            stripe_width = page.chunks[1]
            if (
                page.is_tiled and
                (
                    smallest_stripe_width is None or
                    smallest_stripe_width > stripe_width
                )
            ):
                smallest_stripe_width = stripe_width
        return smallest_stripe_width

    def _create_page(
        self,
        series: int,
        level: int,
        page: int,
    ) -> NdpiPage:
        """Create a new page from TiffPage.

        Parameters
        ----------
        series: int
            Series of page.
        level: int
            Level of page.
        page: int
            Page to use.

        Returns
        ----------
        NdpiLevel
            Created level.
        """
        tiff_page = self._tiff_file.series[series].levels[level].pages[page]
        assert isinstance(tiff_page, TiffPage)
        if tiff_page.is_tiled:  # Striped ndpi page
            return NdpiStripedPage(
                tiff_page,
                self._fh,
                self.base_size,
                self.tile_size,
                self._jpeg
            )
        if series == self._level_series_index:  # Single frame, force tiling
            return NdpiOneFramePage(
                tiff_page,
                self._fh,
                self.base_size,
                self.tile_size,
                self._jpeg
            )
        return NdpiPage(
            tiff_page,
            self._fh,
            self._jpeg
        )  # Single frame, do not tile



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import Dict, Optional, Tuple, Type, Union

from tifffile import TiffFile, TiffFileError

from opentile.common import Tiler
from opentile.histech_tiff_tiler import HistechTiffTiler
from opentile.ndpi_tiler import NdpiTiler
from opentile.philips_tiff_tiler import PhilipsTiffTiler
from opentile.svs_tiler import SvsTiler
from opentile.turbojpeg_patch import find_turbojpeg_path


class OpenTile:
    @staticmethod
    def get_tiler(
        filepath: Union[str, Path]
    ) -> Tuple[Optional[str], Optional[Type[Tiler]]]:
        """Return tiler that supports the tiff file in filepath, or None if
        not supported"""
        tilers: Dict[str, Type[Tiler]] = {
            'ndpi': NdpiTiler,
            'svs': SvsTiler,
            'phillips_tiff': PhilipsTiffTiler,
            '3dhistech tiff': HistechTiffTiler
        }
        try:
            tiff_file = TiffFile(filepath)
            return next(
                (tiler_name, tiler) for (tiler_name, tiler) in tilers.items()
                if tiler.supported(tiff_file)
            )
        except (TiffFileError, StopIteration):
            return None, None

    @classmethod
    def detect_format(
        cls,
        filepath: Union[str, Path]
    ) -> Optional[str]:
        format, supported_tiler = cls.get_tiler(filepath)
        return format

    @classmethod
    def open(
        cls,
        filepath: Union[str, Path],
        tile_size: int = 512,
    ) -> Tiler:
        """Return a file type specific tiler for tiff file in filepath.
        Tile size and turbo jpeg path are optional but required for some file
        types.

        Parameters
        ----------
        filepath: Union[str, Path]
            Path to tiff file.
        tile_size: int = 512
            Tile size for creating tiles, if needed for file format.
        """
        format, supported_tiler = cls.get_tiler(filepath)
        if supported_tiler is NdpiTiler:
            return NdpiTiler(
                filepath,
                tile_size,
                find_turbojpeg_path()
            )

        if supported_tiler is SvsTiler:
            return SvsTiler(
                filepath,
                find_turbojpeg_path()
            )

        if supported_tiler is PhilipsTiffTiler:
            return PhilipsTiffTiler(
                filepath,
                find_turbojpeg_path()
            )

        if supported_tiler is HistechTiffTiler:
            return HistechTiffTiler(
                filepath
            )

        raise NotImplementedError('Non supported tiff file')



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math
import threading
from abc import ABCMeta, abstractclassmethod, abstractmethod
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import numpy as np
from tifffile.tifffile import (COMPRESSION, PHOTOMETRIC, FileHandle, TiffFile,
                               TiffPage, TiffPageSeries, TiffTags)

from opentile.geometry import Point, Region, Size, SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata


class LockableFileHandle:
    """A lockable file handle for reading frames."""
    def __init__(self, fh: FileHandle):
        self._fh = fh
        self._lock = threading.Lock()

    def __str__(self) -> str:
        return f"{type(self).__name__} for FileHandle {self._fh}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._fh})"

    @property
    def filepath(self) -> Path:
        return Path(self._fh.path)

    def read(self, offset: int, bytecount: int) -> bytes:
        """Return bytes from file handle.

        Parameters
        ----------
        offset: int
            Offset in bytes.
        bytecount: int
            Length in bytes.

        Returns
        ----------
        bytes
            Requested bytes.
        """
        with self._lock:
            self._fh.seek(offset)
            data: bytes = self._fh.read(bytecount)
        return data

    def close(self) -> None:
        """Close the file handle"""
        self._fh.close()


class OpenTilePage(metaclass=ABCMeta):
    """Abstract class for reading tiles from TiffPage. Should be inherited to
    support different tiff formats:

    OpenTilePage
        NativeTiledPage - Meta class for pages that are natively tiled
            PhilipsTiffTiledPage - OpenTiledPage for Philips Tiff-page
            SvsTiledPage - OpenTiledPage for Svs Tiff-page
        NdpiPage - Ndpi page that should not be tiled
            NdpiTiledPage - Metaclass for a tiled Ndpi page
                NdpiOneFramePage - Tiled Ndpi page of only one frame
                NdpiStripedPage - Tiled Ndpi page of striped frames

    """
    _pyramid_index: int

    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle
    ):
        """Abstract class for reading tiles from TiffPage.

        Parameters
        ----------
        page: TiffPage
            TiffPage to get tiles from.
        fh: FileHandle
            FileHandle for reading data.
        """
        self._page = page
        self._fh = LockableFileHandle(fh)
        self._image_size = Size(self._page.shape[1], self._page.shape[0])
        if self.page.is_tiled:
            self._tile_size = Size(
                self.page.tilewidth,
                self.page.tilelength
            )
        else:
            self._tile_size = self.image_size
        self._tiled_region = Region(
            position=Point(0, 0),
            size=self.tiled_size - 1
        )

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return f"{type(self).__name__} of page {self._page}"

    @property
    def filepath(self) -> Path:
        return self._fh.filepath

    @property
    def suggested_minimum_chunk_size(self) -> int:
        """Suggested minimum chunk size regarding performance for reading
        multiple tiles (get_tiles())."""
        return 1

    @property
    def compression(self) -> COMPRESSION:
        """Return compression of page."""
        return COMPRESSION(self._page.compression)

    @property
    def photometric_interpretation(self) -> PHOTOMETRIC:
        """Return photometric interpretation, e.g. 'YCBCR" or "RGB"."""
        return PHOTOMETRIC(self._page.photometric)

    @property
    def subsampling(self) -> Optional[Tuple[int, int]]:
        """Return subsampling, or None if only one component."""
        return self._page.subsampling

    @property
    def samples_per_pixel(self) -> int:
        """Return samples per pixel."""
        return self._page.samplesperpixel

    @property
    def page(self) -> TiffPage:
        """Return source TiffPage."""
        return self._page

    @property
    def focal_plane(self) -> float:
        """Return focal plane (in um)."""
        return 0.0

    @property
    def optical_path(self) -> str:
        """Return optical path identifier."""
        # Not sure if optical paths are defined in tiff files...
        return '0'

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel size of the tiles. Returns image size if not tiled page
        """
        return self._tile_size

    @property
    def tiled_size(self) -> Size:
        """The size of the image when tiled."""
        if self.tile_size == Size(0, 0):
            return Size(1, 1)
        return self.image_size.ceil_div(self.tile_size)

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer. Returns 0 for
        images not in pyramidal series."""
        return self._pyramid_index

    @property
    @abstractmethod
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Should return the pixel size in mm/pixel of the page."""
        raise NotImplementedError()

    @abstractmethod
    def get_tile(self, tile_position: Tuple[int, int]) -> bytes:
        """Should return image bytes for tile at tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Should return decoded tile for tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        raise NotImplementedError()

    def get_tiles(
        self,
        tile_positions: Sequence[Tuple[int, int]]
    ) -> List[bytes]:
        """Return list of image bytes for tiles at tile positions.

        Parameters
        ----------
        tile_positions: Sequence[Tuple[int, int]]
            Tile positions to get.

        Returns
        ----------
        List[bytes]
            List of tile bytes.
        """
        return [
            self.get_tile(tile) for tile in tile_positions
        ]

    def get_decoded_tiles(
        self, tile_positions: Sequence[Tuple[int, int]]
    ) -> List[np.ndarray]:
        """Return list of decoded tiles for tiles at tile positions.

        Parameters
        ----------
        tile_positions: Sequence[Tuple[int, int]]
            Tile positions to get.

        Returns
        ----------
        List[np.ndarray]
            List of decoded tiles.
        """
        return [self.get_decoded_tile(tile) for tile in tile_positions]

    def close(self) -> None:
        """Close filehandle."""
        self._fh.close()

    def pretty_str(
        self,
        indent: int = 0,
        depth: Optional[int] = None
    ) -> str:
        return str(self)

    @property
    def tiled_region(self) -> Region:
        """Tile region covering the OpenTilePage."""
        return self._tiled_region

    def valid_tiles(self, region: Region) -> bool:
        """Check if tile region is inside tile geometry and z coordinate and
        optical path exists.

        Parameters
        ----------
        region: Region
            Tile region.
        """
        return region.is_inside(self.tiled_region)

    def _read_frame(self, index: int) -> bytes:
        """Read frame bytes at index from file.

        Parameters
        ----------
        index: int
            Index of frame to read.

        Returns
        ----------
        bytes
            Frame bytes.
        """
        return self._fh.read(
            self._page.dataoffsets[index],
            self._page.databytecounts[index]
        )

    def _check_if_tile_inside_image(self, tile_position: Point) -> bool:
        """Return true if tile position is inside tiled image."""
        return (
            tile_position.x < self.tiled_size.width and
            tile_position.y < self.tiled_size.height
        )

    @staticmethod
    def _get_value_from_tiff_tags(
        tiff_tags: TiffTags,
        value_name: str
    ) -> Optional[str]:
        for tag in tiff_tags:
            if tag.name == value_name:
                return str(tag.value)
        return None

    def _calculate_pyramidal_index(
        self,
        base_shape: Size,
    ) -> int:
        return int(math.log2(base_shape.width / self.image_size.width))

    def _calculate_mpp(
        self,
        base_mpp: SizeMm
    ) -> SizeMm:
        return base_mpp * pow(2, self.pyramid_index)


class NativeTiledPage(OpenTilePage, metaclass=ABCMeta):
    """Meta class for pages that are natively tiled (e.g. not ndpi)"""
    def get_tile(
        self,
        tile_position: Tuple[int, int]
    ) -> bytes:
        """Return image bytes for tile at tile position.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        tile_point = Point.from_tuple(tile_position)
        frame_index = self._tile_point_to_frame_index(tile_point)
        tile = self._read_frame(frame_index)
        if self.page.jpegtables is not None:
            tile = Jpeg.add_jpeg_tables(tile, self.page.jpegtables)
        return tile

    def get_decoded_tile(self, tile_position: Tuple[int, int]) -> np.ndarray:
        """Return decoded tile for tile position. Returns a white tile if tile
        is outside of image.

        Parameters
        ----------
        tile_position: Tuple[int, int]
            Tile position to get.

        Returns
        ----------
        bytes
            Produced tile at position.
        """
        tile_point = Point.from_tuple(tile_position)
        if not self._check_if_tile_inside_image(tile_point):
            return np.full(
                self.tile_size.to_tuple() + (3,),
                255,
                dtype=np.dtype(np.uint8)
            )

        shape: tuple[int, int, int, int]
        frame = self.get_tile(tile_position)
        frame_index = self._tile_point_to_frame_index(tile_point)
        data, _, shape = self.page.decode(frame, frame_index)
        assert isinstance(data, np.ndarray)
        data.shape = shape[1:]
        return data

    def _tile_point_to_frame_index(
        self,
        tile_point: Point
    ) -> int:
        """Return linear frame index for tile position."""
        frame_index = tile_point.y * self.tiled_size.width + tile_point.x
        return frame_index


class Tiler(metaclass=ABCMeta):
    """Abstract class for reading pages from TiffFile."""
    _level_series_index: int = 0
    _overview_series_index: Optional[int] = None
    _label_series_index: Optional[int] = None
    _icc_profile: Optional[bytes] = None

    def __init__(self, filepath: Path):
        """Abstract class for reading pages from TiffFile.

        Parameters
        ----------
        filepath: Path
            Filepath to a TiffFile.
        """
        self._tiff_file = TiffFile(filepath)
        base_page = self.series[self._level_series_index].pages[0]
        assert isinstance(base_page, TiffPage)
        self._base_page = base_page
        self._base_size = Size(
            self.base_page.shape[1],
            self.base_page.shape[0]
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def metadata(self) -> Metadata:
        raise NotImplementedError()

    @property
    def base_page(self) -> TiffPage:
        """Return base pyramid level in pyramid series."""
        return self._base_page

    @property
    def base_size(self) -> Size:
        """Return size of base pyramid level in pyramid series."""
        return self._base_size

    @property
    def series(self) -> List[TiffPageSeries]:
        """Return contained TiffPageSeries."""
        return self._tiff_file.series

    @property
    def levels(self) -> List[OpenTilePage]:
        """Return list of pyramid level OpenTilePages."""
        if self._level_series_index is None:
            return []
        return [
            self.get_level(level_index, page_index)
            for level_index, level
            in enumerate(self.series[self._level_series_index].levels)
            for page_index, page in enumerate(level.pages)
        ]

    @property
    def labels(self) -> List[OpenTilePage]:
        """Return list of label OpenTilePage."""
        if self._label_series_index is None:
            return []
        return [
            self.get_label(page_index)
            for page_index, page
            in enumerate(self.series[self._label_series_index].pages)
        ]

    @property
    def overviews(self) -> List[OpenTilePage]:
        """Return list of overview OpenTilePage."""
        if self._overview_series_index is None:
            return []
        return [
            self.get_overview(page_index)
            for page_index, page
            in enumerate(self.series[self._overview_series_index].pages)
        ]

    @property
    def icc_profile(self) -> Optional[bytes]:
        """Return icc profile if found in file."""
        return self._icc_profile

    @abstractmethod
    def get_page(self, series: int, level: int, page: int) -> OpenTilePage:
        """Should return a OpenTilePage for series, level, page in file."""
        raise NotImplementedError()

    @abstractclassmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        raise NotImplementedError()

    def close(self) -> None:
        """CLose tiff-file."""
        self._tiff_file.close()

    def get_tile(
        self,
        series: int,
        level: int,
        page: int,
        tile_position: Tuple[int, int]
    ) -> bytes:
        """Return list of image bytes for tiles at tile positions.

        Parameters
        ----------
        series: int
            Series of page to get tile from.
        level: int
            Level of page to get tile from.
        page: int
            Page to get tile from.
        tile_position: Tuple[int, int]
            Position of tile to get.

        Returns
        ----------
        bytes
            Tile at position.
        """
        tiled_page = self.get_page(series, level, page)
        return tiled_page.get_tile(tile_position)

    def get_level(
        self,
        level: int,
        page: int = 0
    ) -> OpenTilePage:
        """Return OpenTilePage for level in pyramid series.

        Parameters
        ----------
        level: int
            Level to get.
        page: int
            Index of page to get.

        Returns
        ----------
        OpenTilePage
            Level OpenTilePage.
        """
        return self.get_page(self._level_series_index, level, page)

    def get_label(
        self,
        page: int = 0
    ) -> OpenTilePage:
        """Return OpenTilePage for label in label series.

        Parameters
        ----------
        page: int
            Index of page to get.

        Returns
        ----------
        OpenTilePage
            Label OpenTilePage.
        """
        if self._label_series_index is None:
            raise ValueError("No label detected in file")
        return self.get_page(self._label_series_index, 0, page)

    def get_overview(
        self,
        page: int = 0
    ) -> OpenTilePage:
        """Return OpenTilePage for overview in overview series.

        Parameters
        ----------
        page: int
            Index of page to get.

        Returns
        ----------
        OpenTilePage
            Overview OpenTilePage.
        """
        if self._overview_series_index is None:
            raise ValueError("No overview detected in file")
        return self.get_page(self._overview_series_index, 0, page)

    def _get_tiff_page(self, series: int, level: int, page: int) -> TiffPage:
        """Return TiffPage for series, level, page."""
        tiff_page = self.series[series].levels[level].pages[page]
        assert isinstance(tiff_page, TiffPage)
        return tiff_page



#    Copyright 2022 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from pathlib import Path
from typing import Dict, Optional, Tuple, Union

from tifffile.tifffile import (PHOTOMETRIC, FileHandle, TiffFile, TiffPage,
                               TiffPageSeries)

from opentile.common import NativeTiledPage, Tiler
from opentile.geometry import Size, SizeMm
from opentile.jpeg import Jpeg
from opentile.metadata import Metadata


class HistechTiffTiledPage(NativeTiledPage):
    def __init__(
        self,
        page: TiffPage,
        fh: FileHandle,
        base_shape: Size
    ):
        """OpenTiledPage for 3DHistech Tiff-page.

        Parameters
        ----------
        page: TiffPage
            TiffPage defining the page.
        fh: FileHandler
            Filehandler to read data from.
        base_shape: Size
            Size of base level in pyramid.
        """
        super().__init__(page, fh)
        self._base_shape = base_shape
        self._pyramid_index = self._calculate_pyramidal_index(self._base_shape)
        self._mpp = self._get_mpp_from_page()

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self._page}, {self._fh}, "
            f"{self._base_shape})"
        )

    @property
    def pixel_spacing(self) -> SizeMm:
        """Return pixel spacing in mm per pixel."""
        return self.mpp * 1000

    @property
    def mpp(self) -> SizeMm:
        """Return pixel spacing in um per pixel."""
        return self._mpp

    @property
    def photometric_interpretation(self) -> PHOTOMETRIC:
        photometric_interpretation = PHOTOMETRIC(self._page.photometric)
        if photometric_interpretation == PHOTOMETRIC.PALETTE:
            return PHOTOMETRIC.MINISBLACK
        return photometric_interpretation

    def _get_mpp_from_page(self) -> SizeMm:
        items_split = self._page.description.split('|')
        header = items_split.pop(0)
        items = {
            key: value
            for (key, value) in (
                item.split(' = ', 1) for item in items_split
            )
        }
        return SizeMm(
            float(items['3dh_PixelSizeX']),
            float(items['3dh_PixelSizeY'])
        ) / 1000 / 1000


class HistechTiffTiler(Tiler):
    def __init__(
        self,
        filepath: Union[str, Path],
        turbo_path: Optional[Union[str, Path]] = None
    ):
        """Tiler for 3DHistech tiff file.

        Parameters
        ----------
        filepath: Union[str, Path]
            Filepath to a 3DHistech-TiffFile.
        """
        super().__init__(Path(filepath))
        self._fh = self._tiff_file.filehandle

        self._turbo_path = turbo_path
        self._jpeg = Jpeg(self._turbo_path)

        self._level_series_index = 0
        for series_index, series in enumerate(self.series):
            if self.is_label(series):
                self._label_series_index = series_index
            elif self.is_overview(series):
                self._overview_series_index = series_index
        self._pages: Dict[Tuple[int, int, int], HistechTiffTiledPage] = {}

    @property
    def metadata(self) -> Metadata:
        """No known metadata for 3DHistech tiff files."""
        return Metadata()

    @classmethod
    def supported(cls, tiff_file: TiffFile) -> bool:
        return '3dh_PixelSizeX' in tiff_file.pages.first.description

    def get_page(
        self,
        series: int,
        level: int,
        page: int = 0
    ) -> HistechTiffTiledPage:
        """Return PhilipsTiffTiledPage for series, level, page."""
        if not (series, level, page) in self._pages:
            self._pages[series, level, page] = HistechTiffTiledPage(
                self._get_tiff_page(series, level, page),
                self._fh,
                self.base_size
            )
        return self._pages[series, level, page]

    @staticmethod
    def is_overview(series: TiffPageSeries) -> bool:
        """Return true if series is a overview series."""
        return False

    @staticmethod
    def is_label(series: TiffPageSeries) -> bool:
        """Return true if series is a label series."""
        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opentile']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.1,<10.0.0',
 'PyTurboJPEG>=1.6.1,<2.0.0',
 'imagecodecs>=2022.9.26,<2023.0.0',
 'numpy>=1.22.0,<2.0.0',
 'tifffile>=2022.5.4,<2023.0.0']

setup_kwargs = {
    'name': 'opentile',
    'version': '0.4.2',
    'description': 'Read tiles from wsi-TIFF files',
    'long_description': "# *opentile*\n\n*opentile* is a Python library for reading tiles from wsi tiff files. The aims of the proect are:\n\n- Allow compressed tiles to be losslessly read from wsi tiffs using 2D coordinates (tile position x, y).\n- Provide unified interface for relevant metadata.\n- Support all file formats supported by tifffile that has a non-overlapping tile structure.\n\nCrrently supported file formats are listed and described under *Supported file formats*.\n\n## Installing *opentile*\n\n*opentile* is available on PyPI:\n\n```console\n$ pip install opentile\n```\n\n## Important note\n\nPlease note that this is an early release and the API is not frozen yet. Function names and functionality is prone to change.\n\n## Requirements\n\n*opentile* requires python >=3.7 and uses numpy, Pillow, TiffFile and PyTurboJPEG (with lib-turbojpeg >= 2.1 ).\n\n## Limitations\n\nFiles with z-stacks are currently not fully supported for all formats.\n\n## Supported file formats\n\nThe following description of the workings of the supported file formats does not include the additional specifics for each format that is handled by tifffile. Additional formats supported by tifffile and that have non-overlapping tile layout are likely to be added in future release.\n\n***Hamamatsu Ndpi***\nThe Ndpi-format uses non-rectangular tile size typically 8 pixels high, i.e. stripes. To form tiles, first multiple stripes are concatenated to form a frame covering the tile region. Second, if the stripes are longer than the tile width, the tile is croped out of the frame. The concatenation and crop transformations are performed losslessly.\n\nA ndpi-file can also contain non-tiled images. If these are part of a pyramidal series, *opentile* tiles the image.\n\n***Philips tiff***\nThe Philips tiff-format allows tiles to be sparse, i.e. missing. For such tiles, *opentile* instead provides a blank (currently white) tile image using the same jpeg header as the rest of the image.\n\n***Aperio svs***\nSome Asperio svs-files have corrupt tile data at edges of non-base pyramidal levels. This is observed as tiles with 0-byte length and tiles with incorrect pixel data. *opentile* detects such corruption and instead returns downscaled image data from lower levels. Associated images (label, overview) are currently not handled correctly.\n\n***3DHistech tiff***\nOnly the pyramidal levels are supported (not overviews or labels).\n\n## Metadata\n\nFile metadata can be accessed through the `metadata`-property of a tiler. Depending on file format and content, the following metadata is avaiable:\n\n- Magnification\n- Scanner manufacturer\n- Scanner model\n- Scanner software versions\n- Scanner serial number\n- Aquisition datetime\n\n## Basic usage\n\n***Load a Ndpi-file using tile size (1024, 1024) pixels.***\n\n```python\nfrom opentile import OpenTile\ntile_size = (1024, 1024)\nturbo_path = 'C:/libjpeg-turbo64/bin/turbojpeg.dll'\ntiler = OpenTile.open(path_to_ndpi_file, tile_size, turbo_path)\n```\n\n***Get rectangular tile at level 0 and position x=0, y=0.***\n\n```python\ntile = tiler.get_tile(0, (0, 0))\n```\n\n***Close the tiler object.***\n\n```python\ntiler.close()\n```\n\n***Usage as context manager***\n\nThe tiler can also be used as context manager:\n\n```python\nfrom opentile import OpenTile\ntile_size = (1024, 1024)\nturbo_path = 'C:/libjpeg-turbo64/bin/turbojpeg.dll'\nwith OpenTile.open(path_to_ndpi_file, tile_size, turbo_path) as tiler:\n    tile = tiler.get_tile(0, (0, 0))\n```\n\n## Setup environment for development\n\nRequires poetry and pytest and pytest-watch installed in the virtual environment.\n\n```console\n$ git clone https://github.com/imi-bigpicture/opentile.git\n$ poetry install\n```\n\nBy default the tests looks for slides in 'tests/testdata'. This can be overriden by setting the OPENTILE_TESTDIR environment variable. The script 'tests/download_test_images.py' can be used to download publically available [openslide testdata](https://openslide.cs.cmu.edu/download/openslide-testdata/) into the set testdata folder:\n\n```console\n$ python tests/download_test_images.py\n```\n\nThe test data used for philips tiff is currently not publically available as we dont have permission to share them. If you have slides in philips tiff format that can be freely shared we would be happy to use them instead.\n\nTo watch unit tests use:\n\n```console\n$ poetry run pytest-watch -- -m unittest\n```\n\n## Other TIFF python tools\n\n- [tifffile](https://github.com/cgohlke/tifffile)\n- [tiffslide](https://github.com/bayer-science-for-a-better-life/tiffslide)\n\n## Contributing\n\nWe welcome any contributions to help improve this tool for the WSI community!\n\nWe recommend first creating an issue before creating potential contributions to check that the contribution is in line with the goals of the project. To submit your contribution, please issue a pull request on the imi-bigpicture/opentile repository with your changes for review.\n\nOur aim is to provide constructive and positive code reviews for all submissions. The project relies on gradual typing and roughly follows PEP8. However, we are not dogmatic. Most important is that the code is easy to read and understand.\n\n## Acknowledgement\n\n*opentile*: Copyright 2021 Sectra AB, licensed under Apache 2.0.\n\nThis project is part of a project that has received funding from the Innovative Medicines Initiative 2 Joint Undertaking under grant agreement No 945358. This Joint Undertaking receives support from the European Union’s Horizon 2020 research and innovation programme and EFPIA. IMI website: www.imi.europa.eu",
    'author': 'Erik O Gabrielsson',
    'author_email': 'erik.o.gabrielsson@sectra.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/imi-bigpicture/opentile',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
